#include <fxsaber\Virtual\Order_Base.mqh> // https://www.mql5.com/ru/code/22577

#include "Object_Trend.mqh"
#include "Object_Arrow.mqh"

#ifndef ArraySortStruct2
  #define ArraySortStruct2_Delete

  // https://www.mql5.com/ru/forum/170952/page222#comment_40289584
  //          .
  #define ArraySortStruct2_Define(FIELD)                               \
  namespace SortOnField_##FIELD                                        \
  {                                                                    \
    class SORT2                                                        \
    {                                                                  \
    private:                                                           \
      template <typename T, typename T2>                               \
      static void Sort( T &Array[], const T2& )                        \
      {                                                                \
        T2 SortIndex[][2];                                             \
                                                                       \
        const int Size = ::ArrayResize(SortIndex, ::ArraySize(Array)); \
                                                                       \
        for (int i = Size - 1; i >= 0; i--)                            \
        {                                                              \
          SortIndex[i][0] = (T2)Array[i].##FIELD;                      \
          SortIndex[i][1] = (T2)i;                                     \
        }                                                              \
                                                                       \
        ::ArraySort(SortIndex);                                        \
                                                                       \
        T Sort_Array[];                                                \
                                                                       \
        for (int i = ::ArrayResize(Sort_Array, Size) - 1; i >= 0; i--) \
          Sort_Array[i] = Array[(int)SortIndex[i][1]];                 \
                                                                       \
        ::ArraySwap(Sort_Array, Array);                                \
                                                                       \
        return;                                                        \
      }                                                                \
                                                                       \
    public:                                                            \
      template <typename T>                                            \
      static void Sort( T &Array[] )                                   \
      {                                                                \
        if (::ArraySize(Array))                                        \
          SORT2::Sort(Array, Array[0].##FIELD);                        \
                                                                       \
        return;                                                        \
      }                                                                \
    };                                                                 \
  }

  #define ArraySortStruct2(ARRAY, FIELD) SortOnField_##FIELD::SORT2::Sort(ARRAY)
#endif // #ifndef ArraySortStruct2

ArraySortStruct2_Define(Ticket)
ArraySortStruct2_Define(TimeClose)

#define RESERVE 10000

class OBJECTSTRADE
{
private:
  OBJECT_TREND Objects_Trend[];
  OBJECT_ARROW Objects_Arrow[];

  int Filter;

  //      
  template <typename T>
  static int AddElement( T &Array[], const T &Value, const int Reserve = 0 ) // https://www.mql5.com/ru/forum/1111/page2085#comment_6019715
  {
    const int Size = ::ArraySize(Array);
    const int Res = ::ArrayResize(Array, Size + 1, Reserve);

    Array[Size] = Value;

    return(Res);
  }

  int GetTicketPos( const long Ticket ) const
  {
    int Left = -1;
    int Right = ::ArraySize(this.Objects_Arrow);

    while (Left < Right - 1)
    {
      const int Middle = (Left + Right) >> 1;

      if (this.Objects_Arrow[Middle].Ticket <= Ticket)
        Left = Middle;
      else
        Right = Middle;
    }

    return(Right - 1);
  }

  void TrendToOrder( const OBJECT_TREND &Object_Trend, ORDER_BASE &Order ) const
  {
    ENUM_DEAL_REASON eCloseReason;
    double dProfit;
    string sSymbol;

    Object_Trend.Get(eCloseReason, dProfit, sSymbol);

    const string sCommentOpen = this.Objects_Arrow[this.GetTicketPos(Object_Trend.Ticket)].Text.Get();

    const int PosClose = this.GetTicketPos(Object_Trend.TicketClose);
    string sComment = this.Objects_Arrow[PosClose].Text.Get();

    ENUM_ORDER_TYPE eType;
    double dLots;

    this.Objects_Arrow[PosClose].Get(eType, dLots);

    double dClosePriceRequest = Object_Trend.PriceClose;

    double dTakeProfit = 0;
    double dStopLoss = 0;

    if (eCloseReason != DEAL_REASON_CLIENT)
    {
      dClosePriceRequest = (double)::StringSubstr(sComment, ::StringFind(sComment, (eCloseReason == DEAL_REASON_TP) ? "tp " : "sl ") + 3);

      dTakeProfit = (eCloseReason == DEAL_REASON_TP) ? dClosePriceRequest : 0;
      dStopLoss = (eCloseReason == DEAL_REASON_SL) ? dClosePriceRequest : 0;
    }

    if (sCommentOpen != "")
      sComment = sCommentOpen;

    Order.Set(Object_Trend.TicketClose, eType, dLots, sSymbol, sComment, Object_Trend.PriceOpen, Object_Trend.TimeOpen * 1000,
              dStopLoss, dTakeProfit, Object_Trend.PriceClose, Object_Trend.TimeClose * 1000, 0, dProfit, 0, 0,
              DEAL_REASON_CLIENT, eCloseReason, 0, dClosePriceRequest);

    return;
  }

  static bool IsTester( const OBJECT_TREND &Object_Trend )
  {
    return(Object_Trend.Ticket < 100000);
  }

  static bool IsSignal( const OBJECT_TREND &Object_Trend )
  {
    return(Object_Trend.Ticket > 600000000);
  }

  void SetFilter( const OBJECT_TREND &Object_Trend )
  {
    if (!this.Filter)
    {
      if (OBJECTSTRADE::IsTester(Object_Trend))
        this.Filter = -1;
      else if (OBJECTSTRADE::IsSignal(Object_Trend))
        this.Filter = 1;
    }
    else if ((this.Filter == -1) && OBJECTSTRADE::IsSignal(Object_Trend))
      this.Filter = 1;

    return;
  }

  bool IsFilter( const OBJECT_TREND &Object_Trend ) const
  {
    return((!this.Filter && !OBJECTSTRADE::IsTester(Object_Trend) && !OBJECTSTRADE::IsSignal(Object_Trend)) ||
           ((this.Filter == -1) && OBJECTSTRADE::IsTester(Object_Trend)) ||
           ((this.Filter == 1) && OBJECTSTRADE::IsSignal(Object_Trend)));
  }

public:
  OBJECTSTRADE( void ) : Filter(0)
  {
  }

  OBJECTSTRADE( const long Chart ) : Filter(0)
  {
    this.Set(Chart);
  }

  int Set( const long Chart = 0, const bool Sort = true ) // Chart = -1 - All Charts.
  {
    if (Chart >= 0)
      for (int i = ::ObjectsTotal(Chart, 0) - 1; !::IsStopped() && (i >= 0); i--)
      {
        static OBJECT_TREND Object_Trend;
        static OBJECT_ARROW Object_Arrow;

        const string Name = ::ObjectName(Chart, i, 0);

        switch ((ENUM_OBJECT)ObjectGetInteger(Chart, Name, OBJPROP_TYPE))
        {
          case OBJ_TREND:
            if (Object_Trend.Set(Name, Chart))
            {
              OBJECTSTRADE::AddElement(this.Objects_Trend, Object_Trend, RESERVE);

              this.SetFilter(Object_Trend);
            }
            break;
          case OBJ_ARROW_BUY:
          case OBJ_ARROW_SELL:
            if (Object_Arrow.Set(Name, Chart))
              OBJECTSTRADE::AddElement(this.Objects_Arrow, Object_Arrow, RESERVE);
        }
      }
    else
      for (long lChart = ::ChartFirst(); !::IsStopped() && (lChart != -1); lChart = ::ChartNext(lChart))
        this.Set(lChart, false);

    if (Sort && !::IsStopped())
      this.Sort();

    return(::ArraySize(this.Objects_Trend));
  }

  void Sort( void )
  {
    ArraySortStruct2(this.Objects_Trend, TimeClose);

    ArraySortStruct2(this.Objects_Arrow, Ticket);

    return;
  }

  int Get( ORDER_BASE &Orders[] ) const
  {
    const int Size = ::ArrayResize(Orders, ::ArraySize(this.Objects_Trend));
    int Amount = 0;

    for (int i = 0; i < Size; i++)
      if (this.IsFilter(this.Objects_Trend[i]))
        this.TrendToOrder(this.Objects_Trend[i], Orders[Amount++]);

    return(::ArrayResize(Orders, Amount));
  }

  void Clear( void )
  {
    ::ArrayFree(this.Objects_Trend);
    ::ArrayFree(this.Objects_Arrow);

    this.Filter = 0;

    return;
  }

  void SetFilter( const int Value )
  {
    this.Filter = Value ? ((Value > 0) ? 1 : -1) : 0;

    return;
  }

  int GetFilter( void ) const
  {
    return(this.Filter);
  }

  string ToString( void ) const
  {
    return(this.Filter ? ((this.Filter > 0) ? "Signal" : "Tester") : "Account");
  }
};

#undef RESERVE

#ifdef ArraySortStruct2_Delete
  #undef ArraySortStruct2
  #undef ArraySortStruct2_Define
  #undef ArraySortStruct2_Delete
#endif // #ifdef ArraySortStruct_Delete
